import inspect
from abc import abstractmethod
from typing import (
    Any,
    ClassVar,
    List,
    Literal,
    LiteralString,
    TypeVar,
    get_origin,
)

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from .utils import FHIRType, get_source_type

TUrl = TypeVar("TUrl", bound=LiteralString)
TFhirType = TypeVar("TFhirType", bound=FHIRType)
TPythonType = TypeVar("TPythonType")


# resource.extension["itemControl"].valueCodeableConcept.coding["tiro"].code

DiscriminatorType = Literal["value", "exists", "type"]
TDefaultElement = TypeVar("TDefaultElement")


class BaseElementArray[TDefaultElement](List[TDefaultElement]):
    """A collection of elements that can be sliced and named using a discriminator."""

    @classmethod
    def get_slice_annotations(cls) -> dict[str, type]:
        return {
            slice_name if slice_name != "_" else "@default": slice_type
            for slice_name, slice_type in inspect.get_annotations(cls).items()
            if get_origin(slice_type) is not ClassVar
        }

    @classmethod
    def get_schema_for_slices(cls, handler: GetCoreSchemaHandler):
        """Generate a schema for each slice.

        Args:
            handler (GetCoreSchemaHandler): The handler to generate the schema

        Yields:
            tuple[str, CoreSchema]: The name of the slice and the schema
        """
        for slice_name, slice_type in cls.get_slice_annotations().items():
            source_types = list(get_source_type(slice_type))
            match len(source_types):
                case 0:
                    raise TypeError(f"Expected a source type, got none for {slice_name}")
                case 1:
                    yield slice_name, handler(source_types[0])
                case _:
                    yield slice_name, core_schema.union_schema([handler(source_type) for source_type in source_types])

    @classmethod
    @abstractmethod
    def discriminator(cls, value: Any) -> str | None:
        """Get the discriminator value for a given value."""
        ...

    @classmethod
    def check_slice_annotations(cls):
        """Check the slice annotations.
        This is a general check for slice annotations. Subclass should extend this method
        to provide specific checks for their own slice annotations.
        """
        for slice_name, slice_type in cls.get_slice_annotations().items():
            source_types = list(get_source_type(slice_type))
            if len(source_types) == 0:
                raise TypeError(f"Expected a source type, got {source_types}")

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler):
        """Get the Pydantic core schema for the element array."""
        cls.check_slice_annotations()
        choices: dict[str, CoreSchema] = dict(cls.get_schema_for_slices(handler))
        schema = core_schema.list_schema(
            core_schema.tagged_union_schema(choices=choices, discriminator=cls.discriminator)
        )
        # TODO add after validators for cardinality of each slice
        return core_schema.json_or_python_schema(
            core_schema.no_info_after_validator_function(
                cls,
                schema,
            ),
            core_schema.union_schema(
                [
                    core_schema.is_instance_schema(cls),
                    core_schema.no_info_after_validator_function(cls, schema),
                ]
            ),
        )


if __name__ == "__main__":
    pass



from fhir_slicing.base import BaseModel
from fhir_slicing.coding import BaseCodingArray, GeneralCoding, LOINCCoding, SCTCoding
from fhir_slicing.slice import OptionalSlice, Slice, SliceList, slice


def test_multi_coding_concepts():
    class CodingArray(BaseCodingArray):
        sct: Slice[SCTCoding] = slice(1, 1)
        loinc: OptionalSlice[LOINCCoding] = slice(0, 1)
        _: SliceList[GeneralCoding] = slice(0, "*")

    class CodeableConcept(BaseModel):
        coding: CodingArray
        text: str | None = None

    concept = CodeableConcept.model_validate(
        {
            "coding": [
                {"system": "http://snomed.info/sct", "code": "123456", "display": "Test"},
                {"system": "http://loinc.org", "code": "123456", "display": "Test"},
                {"system": "http://other.org", "code": "123456", "display": "Test"},
            ],
            "text": "Test",
        }
    )

    assert concept.coding.sct.system == "http://snomed.info/sct"
    assert concept.coding.loinc is not None, "Expected loinc to be present"
    assert concept.coding.loinc.system == "http://loinc.org"



from typing import Any, Mapping, Type

from .base import BaseModel
from .coding import BaseCoding
from .element_array import BaseElementArray
from .utils import get_source_type, get_value_from_literal


class UsageCoding[TSystem: str, TCode: str, TVersion: str | None](BaseCoding[TSystem]):
    code: TCode
    version: TVersion | None = None

    def __eq__(self, other: object):
        if not isinstance(other, UsageCoding):
            return False
        return hash((self.system, self.code, self.version)) == hash((other.system, other.code, other.version))

    @classmethod
    def get_code(cls):
        """Get the coding representation"""
        system: TSystem | None = cls.get_system()
        code: TCode | None = get_value_from_literal(cls.model_fields["code"].annotation)  # type: ignore[override]
        version: TVersion | None = get_value_from_literal(cls.model_fields["version"].annotation)  # type: ignore[override]
        if system is None or code is None:
            return None
        return UsageCoding(system=system, code=code, version=version)


class BaseUsageContext[TUsageCoding: UsageCoding](BaseModel):
    code: TUsageCoding
    model_config = {"extra": "allow"}

    @classmethod
    def get_code(cls) -> TUsageCoding | None:
        """Get the coding representation"""
        code_model: Type[TUsageCoding] = cls.model_fields["code"].annotation  # type: ignore[override]
        usage_code: TUsageCoding | None = code_model.get_code()  # type: ignore[override]
        return usage_code


class BaseUsageContextArray(BaseElementArray[BaseUsageContext]):
    @classmethod
    def get_code(cls, value: Mapping) -> UsageCoding | None:
        """Return the code of the usage context."""
        return value.get("code", None)

    @classmethod
    def discriminator(cls, value: Any) -> str | None:
        coding = cls.get_code(value)
        for slice_name, slice_annotation in cls.get_slice_annotations().items():
            for source_type in get_source_type(slice_annotation, expect=BaseUsageContext):
                if slice_name == "@default" or source_type.get_code() == coding:
                    return slice_name
        return "@default"



from typing import Any, Literal, Mapping

from fhir_slicing.element_array import BaseElementArray
from fhir_slicing.utils import get_source_type, get_value_from_literal

from .base import BaseModel


class BaseCoding[TSystem: str](BaseModel):
    system: TSystem

    @classmethod
    def get_system(cls) -> TSystem | None:
        """Get the system of the coding"""
        system: TSystem | None = get_value_from_literal(cls.model_fields["system"].annotation)  # type: ignore[override]
        if system is None:
            return None
        assert isinstance(system, str), f"Expected system to be a string, got {system}"
        return system


class GeneralCoding(BaseCoding):
    display: str | None = None
    code: str
    model_config = {"extra": "allow"}


class BaseCodingArray(BaseElementArray[GeneralCoding]):
    @classmethod
    def get_system(cls, value: Mapping) -> str | None:
        """Get the system of the coding"""
        return value.get("system", None)

    @classmethod
    def check_slice_annotations(cls):
        super().check_slice_annotations()

        # check if slices have unique systems
        system_slice_name_dict: dict[str | None, str] = dict()
        for slice_name, slice_annotation in cls.get_slice_annotations().items():
            for source_type in get_source_type(slice_annotation, expect=BaseCoding):
                system = source_type.get_system()
                if system in system_slice_name_dict:
                    existing_slice_name = system_slice_name_dict[system]
                    raise TypeError(
                        f"Duplicate discriminator value {system} for {source_type} with slice name {slice_name}. Slice '{existing_slice_name}' has claimed this system."
                    )
                system_slice_name_dict[system] = slice_name

    @classmethod
    def discriminator(cls, value: Any) -> str | None:
        """Get the discriminator value for a given value."""
        # Ex. sct, SCTCoding
        for slice_name, slice_annotation in cls.get_slice_annotations().items():
            for source_type in get_source_type(slice_annotation, expect=BaseCoding):
                if slice_name == "@default" or source_type.get_system() == cls.get_system(value):
                    return slice_name
        return "@default"


class SCTCoding(BaseCoding):
    system: Literal["http://snomed.info/sct"] = "http://snomed.info/sct"
    code: str
    display: str


class LOINCCoding(BaseCoding):
    system: Literal["http://loinc.org"] = "http://loinc.org"
    code: str
    display: str



from typing import Any, Literal

from pydantic import PositiveInt, TypeAdapter

from fhir_slicing import BaseModel, Slice, SliceList, slice
from fhir_slicing.extension import (
    BaseExtensionArray,
    BaseSimpleExtension,
    GeneralExtension,
)


def test_extension_model_get_url():
    class MyExtension(BaseSimpleExtension[Literal["http://example.com"]]):
        valueString: str

    assert MyExtension.get_url() == "http://example.com"


def test_extension_array_from_extension_list():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"]]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"]]):
        valueString: str

    class ExtensionArray(BaseExtensionArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)
        _: SliceList[GeneralExtension] = slice(0, "*")

    ext_list = [
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionB(url="http://example.com/extension-b", valueString="b"),
        GeneralExtension.model_validate({"url": "http://example.com", "valueInteger": 3}),
    ]

    ext_array = ExtensionArray(ext_list)
    assert ext_array.a == ext_list[:3]
    assert ext_array.b == ext_list[3]
    assert list(ext_array) == ext_list


def test_extension_array_validator():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"]]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"]]):
        valueString: str

    class ExtensionArray(BaseExtensionArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)
        _: SliceList[GeneralExtension] = slice(0, "*")

    ext_list = [
        {"url": "http://example.com", "valueInteger": 5},
        {"url": "http://example.com/extension-a", "valueString": "1"},
        {"url": "http://example.com/extension-a", "valueString": "2"},
        {"url": "http://example.com/extension-a", "valueString": "3"},
        {"url": "http://example.com/extension-b", "valueString": "4"},
    ]

    ext_array = TypeAdapter(ExtensionArray).validate_python(ext_list)

    assert ext_array.a == [
        MyExtensionA(url="http://example.com/extension-a", valueString="1"),
        MyExtensionA(url="http://example.com/extension-a", valueString="2"),
        MyExtensionA(url="http://example.com/extension-a", valueString="3"),
    ]

    assert ext_array.b == MyExtensionB(url="http://example.com/extension-b", valueString="4")

    ext_list_roundtrip = TypeAdapter(ExtensionArray).dump_python(ext_array, mode="python")
    assert ext_list_roundtrip == ext_list


def test_extension_array_ordering_roundtrip():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"]]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"]]):
        valueString: str

    class ExtensionArray(BaseExtensionArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)

    ext_array = ExtensionArray(
        (
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionB(url="http://example.com/extension-b", valueString="b"),
        )
    )

    ext_list = TypeAdapter(ExtensionArray).dump_python(ext_array)

    assert ext_list == [
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-b", "valueString": "b"},
    ]

    ext_array_roundtrip = TypeAdapter(ExtensionArray).validate_python(ext_list)

    assert ext_array_roundtrip == ext_array


def test_patient_use_case():
    class MultipleBirth(BaseSimpleExtension[Literal["http://hl7.org/fhir/StructureDefinition/patient-multipleBirth"]]):
        valueInteger: PositiveInt

    class PatientExtensions(BaseExtensionArray):
        multiple_birth: Slice[MultipleBirth] = slice(1, 1)

    class PatientName(BaseModel):
        text: str
        given: list[str] | None = None
        family: str | None = None
        use: Literal["usual", "official", "temp", "nickname", "anounymous", "old", "maiden"] | None = None

    class Patient(BaseModel):
        extensions: PatientExtensions
        resourceType: Literal["Patient"] = "Patient"
        name: list[PatientName] | None = None

        @property
        def multiple_birth(self):
            return self.extensions.multiple_birth.valueInteger

        @multiple_birth.setter
        def set_multiple_birth(self, value: PositiveInt):
            self.extensions.multiple_birth.valueInteger = value

    patient = Patient.model_validate(
        {
            "resourceType": "Patient",
            "name": [
                {
                    "text": "John Doe",
                    "given": ["John"],
                    "family": "Doe",
                    "use": "official",
                },
            ],
            "extensions": [
                {
                    "url": "http://hl7.org/fhir/StructureDefinition/patient-multipleBirth",
                    "valueInteger": 3,
                }
            ],
        }
    )

    assert patient.extensions.multiple_birth.valueInteger == 3
    assert patient.multiple_birth == 3


def test_blood_pressure_use_case():
    class Quantity(BaseModel):
        value: float
        unit: str

    class Coding(BaseModel):
        system: str
        code: str
        display: str

    class CodeableConcept(BaseModel):
        coding: list[Coding]
        text: str | None = None

    class BloodPressureComponent(BaseModel):
        valueQuantity: Quantity
        code: CodeableConcept

    class BloodPressureComponents(BaseExtensionArray):
        systolic: Slice[BloodPressureComponent] = slice(1, 1)
        diastolic: Slice[BloodPressureComponent] = slice(1, 1)

        @classmethod
        def discriminator(cls, value: Any) -> str:
            code = value.get("code", {}).get("coding", [{}])[0].get("code", None)
            match code:
                case "8480-6":
                    return "systolic"
                case "8462-4":
                    return "diastolic"
                case _:
                    return "@default"

    class BloodPressure(BaseModel):
        resourceType: Literal["Observation"] = "Observation"
        code: CodeableConcept
        components: BloodPressureComponents

        @property
        def systolic(self):
            return self.components.systolic.valueQuantity.value

        @property
        def diastolic(self):
            return self.components.diastolic.valueQuantity.value

    blood_pressure = BloodPressure.model_validate(
        {
            "resourceType": "Observation",
            "code": {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "55284-4",
                        "display": "Blood pressure",
                    }
                ],
                "text": "Blood pressure",
            },
            "components": [
                {
                    "valueQuantity": {"value": 120, "unit": "mm[Hg]"},
                    "code": {
                        "coding": [
                            {
                                "system": "http://loinc.org",
                                "code": "8480-6",
                                "display": "Systolic blood pressure",
                            }
                        ],
                        "text": "Systolic blood pressure",
                    },
                },
                {
                    "valueQuantity": {"value": 80, "unit": "mm[Hg]"},
                    "code": {
                        "coding": [
                            {
                                "system": "http://loinc.org",
                                "code": "8462-4",
                                "display": "Diastolic blood pressure",
                            }
                        ],
                        "text": "Diastolic blood pressure",
                    },
                },
            ],
        }
    )

    assert blood_pressure.components.systolic.valueQuantity.value == 120
    assert blood_pressure.systolic == 120
    assert blood_pressure.components.diastolic.valueQuantity.value == 80
    assert blood_pressure.diastolic == 80



from typing import Annotated, Any, Container, Iterable, List, Literal, Protocol, TypeVar, overload

import annotated_types

TValueType = TypeVar("TValueType", covariant=True)


class ElementArray(Protocol):
    @classmethod
    def discriminator(cls, value: Any) -> str | None:
        """Get the discriminator value for a given value."""
        ...


class BaseSlice:
    def __set_name__(self, owner: ElementArray, name: str):
        self.name = name
        self.discriminator_func = owner.discriminator

    def filter_element(self, element):
        return self.discriminator_func(element) == self.name


class Slice[TValueType](BaseSlice):
    def __get__(self, obj: Iterable, objtype: type[Container] | None = None) -> TValueType:
        try:
            return next(iter(filter(self.filter_element, obj)))
        except StopIteration:
            raise ValueError(f"No value for slice '{self.name}'.")

    def __set__(self, obj: List, value: Any):
        for index, element in enumerate(obj):
            if self.discriminator_func(element) == self.discriminator_func(value):
                obj[index] = value
                return
        raise ValueError("Cannot set value on slice.")


class OptionalSlice[TValueType](BaseSlice):
    def __get__(self, obj: Iterable, objtype: type[Container] | None = None) -> TValueType | None:
        return next(iter(filter(self.filter_element, obj)), None)

    def __set__(self, obj: List, value: Any):
        for index, element in enumerate(obj):
            if self.discriminator_func(element) == self.discriminator_func(value):
                obj[index] = value
                return


class SliceList[TValueType](BaseSlice):
    def __set_name__(self, owner: ElementArray, name: str):
        self.name = name
        self.discriminator_func = owner.discriminator

    def filter_element(self, element):
        return self.discriminator_func(element) == self.name

    def __get__(self, obj: Iterable, objtype: type[Container] | None = None) -> List[TValueType]:
        return [*filter(self.filter_element, obj)]

    def __set__(self, obj: List, value: List):
        raise NotImplementedError("Cannot set value on slice list.")


NonZeroPositiveInt = Annotated[int, annotated_types.Gt(0)]


@overload
def slice(min: Literal[0], max: Literal[1]) -> OptionalSlice: ...
@overload
def slice(min: Literal[1], max: Literal[1]) -> Slice: ...
@overload
def slice(min: NonZeroPositiveInt, max: Literal["*"]) -> SliceList: ...
def slice(min: int, max: int | Literal["*"]):
    match (min, max):
        case (1, 1):
            return Slice()
        case (0, 1):
            return OptionalSlice()
        case (0, "*"):
            return SliceList()
        case _:
            raise ValueError("Invalid slice cardinality.")



from dataclasses import fields
from typing import (
    Any,
    Mapping,
)

from .base import BaseModel
from .element_array import BaseElementArray
from .utils import get_source_type, get_value_from_literal


class BaseExtension[TUrl: str](BaseModel):
    url: TUrl

    @classmethod
    def get_url(cls) -> str:
        """Get the url of the extension"""
        url = get_value_from_literal(cls.model_fields["url"].annotation)
        assert isinstance(url, str), f"Expected url to be a string, got {url}"
        return url


class GeneralExtension(BaseExtension):
    model_config = {"extra": "allow"}


class BaseSimpleExtension[TUrl: str](BaseExtension[TUrl]):
    url: TUrl

    @property
    def value(self):
        value_field_name = next(field.name for field in fields(self) if field.name.startswith("value"))
        return getattr(self, value_field_name)


class BaseExtensionArray(BaseElementArray[BaseExtension]):
    @classmethod
    def get_url(cls, value: Mapping) -> str | None:
        """Get the url of the extension"""
        return value.get("url", None)

    @classmethod
    def discriminator(cls, value: Any) -> str:
        url = cls.get_url(value)
        for slice_name, slice_annotation in cls.get_slice_annotations().items():
            for source_type in get_source_type(slice_annotation, expect=BaseExtension):
                if slice_name == "@default" or source_type.get_url() == url:
                    return slice_name
        return "@default"


if __name__ == "__main__":
    pass
